#include <stdio.h>
#include <math.h>
#include "fft.h"

#define NBITS 2048
#define SAMPLES 32
#define NBINS (NBITS*SAMPLES)

/* Return relative energy (volts squared) in FFT bin */
float
energy(complex val)
{
  /* sum of squares of I and Q components */
  return (val.re*val.re) + (val.im*val.im);
}

/* Clip arg to -100 < x < +100 */
float
clip(float x)
{
  if(x < -100)
    x = -100;
  else if(x > 100)
    x = 100;
  return x;
}

/* Generate the so-called VMSK signal and compute fourier transform
 * 8/31/2000 P. Karn
 */
main()
{
  int bits,bit,samp,i;
  complex samples[NBINS];
  float amplitude, val;
  double totenergy;

  /* easy way to zero all the complex values */
  memset(samples,0,sizeof(samples));

  for(bits=0;bits<NBITS;bits++){
    /* Generate random data bit */
    bit = random() & 1;

    /* If the data bit is zero, emit 15 samples of +1s followed by
     * 17 samples of -1s. Otherwise, emit 17 samples of 1s followed
     * by 15 samples of -1s.
     */
    for(samp=0;samp<SAMPLES/2 - 1;samp++)
      samples[bits*SAMPLES + samp].re = 1;

    if(bit == 1)
      for(;samp<SAMPLES/2 + 1;samp++)
	samples[bits*SAMPLES + samp].re = 1;

    for(;samp<SAMPLES;samp++){
      samples[bits*SAMPLES + samp].re = -1;
    }
  }
#ifdef WINDOW
  /* Apply simple Hanning window */
  for(bits=0;bits<NBINS;bits++)
    samples[bits].re *= sin((M_PI*bits) / NBITS);
#endif

  /* Now take the FFT */
  fftinit();
  fftn(samples,8); /* 4**8 = 65536-point FFT */
  fftswap(samples,65536); /* De-interleave frequency samples */

  /* Compute total energy of signal for normalization */
  totenergy = energy(samples[0])/2; /* Half energy for DC */
  for(i=1;i<NBINS;i++)
    totenergy += energy(samples[i]);

  /* Negative frequencies */
  for(i=NBINS/2;i<NBINS;i++){
    val = clip(10*log10(energy(samples[i])/totenergy));
    printf("%f %f\n",(float)(i-NBINS)/(NBINS/SAMPLES),val);
  }

  /* DC is special case; we have to take half the energy here */
  val = clip(10*log10(energy(samples[0])/(2*totenergy)));
  printf("%f %f\n",0.0,val);

  /* Positive frequencies */
  for(i=1;i<NBINS/2;i++){
    val = clip(10*log10(energy(samples[i])/totenergy));
    printf("%f %f\n",(float)i/(NBINS/SAMPLES),val);
  }
}



