/* Radix 4 DIF FFT butterflies in heavily optimized Pentium assembler;
 * great care taken to maximize pipelining in the FPU
 * August 1995, Phil Karn, KA9Q
 */
#define	ASM
#include "fft.h"

#define	COS(x)	Twiddles(x)
#define SIN(x)	Twiddles+MAXPOINTS(x)	/* MAXPOINTS*sizeof(float)/4 */

.globl Twiddles

#define	Are	0(%eax)
#define	Aim	4(%eax)
#define Bre	8(%eax)
#define	Bim	12(%eax)
#define	Cre	16(%eax)
#define	Cim	20(%eax)
#define	Dre	24(%eax)
#define	Dim	28(%eax)	

/* Radix 4 butterfly for the important special case of adjacent data elements
 * and unity twiddles
 * Defined as fft4(complex[4])
 *
 * Both fft4() and fft4g() swap their middle two results (B&C) so that the
 * final result of the FFT is in bit-reversed order just like the radix-2
 * FFT, instead of in "digit reversed" order.
 */
.globl fft4
fft4:
	pushl	%ebp
	movl	%esp,%ebp
	movl	8(%ebp),%eax	/* eax -> x[0] */

	/* We need Bre-Dre for the imaginaries after we've overwritten the reals */
	flds	Bre
	fsubs	Dre	/* Bre-Dre */

	/* Compute the reals */
	flds	Bre
	fadds	Dre	/* Bre+Dre	Bre-Dre */
	flds	Are
	fadds	Cre	/* Are+Cre	Bre+Dre		Bre-Dre */
	flds	Are
	fsubs	Cre	/* Are-Cre	Are+Cre		Bre+Dre		Bre-Dre */
	flds	Bim
	fsubs	Dim	/* Bim-Dim	Are-Cre		Are+Cre		Bre+Dre		Bre-Dre */
	fxch	%st(3)	/* Bre+Dre	Are-Cre		Are+Cre		Bim-Dim		Bre-Dre */
	fld	%st(0)	/* Bre+Dre	Bre+Dre		Are-Cre		Are+Cre		Bim-Dim		Bre-Dre */
	fadd	%st(3)	/* Are'		Bre+Dre		Are-Cre		Are+Cre		Bim-Dim		Bre-Dre */
	fxch	%st(3)	/* Are+Cre	Bre+Dre		Are-Cre		Are'		Bim-Dim		Bre-Dre */
	fsubp		/* Cre'		Are-Cre		Are'		Bim-Dim		Bre-Dre */
	fxch	%st(1)	/* Are-Cre	Cre'		Are'		Bim-Dim		Bre-Dre */
	fld	%st(0)	/* Are-Cre	Are-Cre		Cre'		Are'		Bim-Dim		Bre-Dre */
	fadd	%st(4)	/* Bre'		Are-Cre		Cre'		Are'		Bim-Dim		Bre-Dre */
	fxch	%st(4)	/* Bim-Dim	Are-Cre		Cre'		Are'		Bre'		Bre-Dre */
	fsubrp		/* Dre'		Cre'		Are'		Bre'		Bre-Dre */
	fxch	%st(2)	/* Are'		Cre'		Dre'		Bre'		Bre-Dre */
	fstps	Are	/* Cre'		Dre'		Bre'		Bre-Dre */
	fstps	Bre	/* Dre'		Bre'		Bre-Dre */
	fstps	Dre	/* Bre'		Bre-Dre */
	fstps	Cre	/* Bre-Dre */

	/* Now do the imaginaries */
	flds	Aim
	fsubs	Cim	/* Aim-Cim	Bre-Dre */
	flds	Aim
	fadds	Cim	/* Aim+Cim	Aim-Cim		Bre-Dre */
	flds	Bim
	fadds	Dim	/* Bim+Dim	Aim+Cim		Aim-Cim		Bre-Dre */	
	fld	%st(3)	/* Bre-Dre	Bim+Dim		Aim+Cim		Aim-Cim		Bre-Dre */
	fadd	%st(3)	/* Dim'		Bim+Dim		Aim+Cim		Aim-Cim		Bre-Dre */
	fxch	%st(2)	/* Aim+Cim	Bim+Dim		Dim'		Aim-Cim		Bre-Dre */
	fld	%st(0)	/* Aim+Cim	Aim+Cim		Bim+Dim		Dim'		Aim-Cim		Bre-Dre */
	fadd	%st(2)	/* Aim'		Aim+Cim		Bim+Dim		Dim'		Aim-Cim		Bre-Dre */
	fxch	%st(2)	/* Bim+Dim	Aim+Cim		Aim'		Dim'		Aim-Cim		Bre-Dre */
	fsubrp		/* Cim'		Aim'		Dim'		Aim-Cim		Bre-Dre */
	fxch	%st(3)	/* Aim-Cim	Aim'		Dim'		Cim'		Bre-Dre */
	fsub	%st(4)	/* Bim'		Aim'		Dim'		Cim'		Bre-Dre */
	fxch	%st(2)	/* Dim'		Aim'		Bim'		Cim'		Bre-Dre */

	fstps	Dim	/* Aim'		Bim'		Cim'		Bre-Dre */
	fstps	Aim	/* Bim'		Cim'		Bre-Dre */
	fstps	Cim	/* Cim'		Bre-Dre */
	fstps	Bim	/* Bre-Dre */
	ffree	%st(0)

	mov	%ebp,%esp
	pop	%ebp
	ret

#undef	Are
#undef	Bre
#undef	Cre
#undef	Dre
#undef	Aim
#undef	Bim
#undef	Cim
#undef	Dim

#define	Are	0(%eax)
#define	Aim	4(%eax)
#define	Bre	0(%ebx)
#define	Bim	4(%ebx)
#define	Cre	0(%ecx)
#define	Cim	4(%ecx)
#define	Dre	0(%edx)
#define	Dim	4(%edx)

/* General radix 4 DIF Butterfly with nonadjacent data elements & nonunity twiddles
 * defined as fft4g(complex *a,int astep,int tstep)
 */
	.align 2
.globl fft4g
fft4g:
	pushl %ebp
	movl %esp,%ebp
	push %ebx
	push %ecx
	push %edx

	movl 8(%ebp),%eax	/* eax -> a */
	shll $3,12(%ebp)	/* astep *= sizeof(complex) */
	movl %eax,%ebx
	addl 12(%ebp),%ebx	/* ebx -> b */
	movl %ebx,%ecx
	addl 12(%ebp),%ecx	/* ecx -> c */
	movl %ecx,%edx
	addl 12(%ebp),%edx	/* edx -> d */	

	/* We need Bre-Dre for the imaginaries after we've overwritten the reals */
	flds	Bre
	fsubs	Dre	/* Bre-Dre */

	/* Compute the reals */
	flds	Bre
	fadds	Dre	/* Bre+Dre	Bre-Dre */
	flds	Are
	fadds	Cre	/* Are+Cre	Bre+Dre		Bre-Dre */
	flds	Are
	fsubs	Cre	/* Are-Cre	Are+Cre		Bre+Dre		Bre-Dre */
	flds	Bim
	fsubs	Dim	/* Bim-Dim	Are-Cre		Are+Cre		Bre+Dre		Bre-Dre */
	fxch	%st(3)	/* Bre+Dre	Are-Cre		Are+Cre		Bim-Dim		Bre-Dre */
	fld	%st(0)	/* Bre+Dre	Bre+Dre		Are-Cre		Are+Cre		Bim-Dim		Bre-Dre */
	fadd	%st(3)	/* Are'		Bre+Dre		Are-Cre		Are+Cre		Bim-Dim		Bre-Dre */
	fxch	%st(3)	/* Are+Cre	Bre+Dre		Are-Cre		Are'		Bim-Dim		Bre-Dre */
	fsubp		/* Cre'		Are-Cre		Are'		Bim-Dim		Bre-Dre */
	fxch	%st(1)	/* Are-Cre	Cre'		Are'		Bim-Dim		Bre-Dre */
	fld	%st(0)	/* Are-Cre	Are-Cre		Cre'		Are'		Bim-Dim		Bre-Dre */
	fadd	%st(4)	/* Bre'		Are-Cre		Cre'		Are'		Bim-Dim		Bre-Dre */
	fxch	%st(4)	/* Bim-Dim	Are-Cre		Cre'		Are'		Bre'		Bre-Dre */
	fsubrp		/* Dre'		Cre'		Are'		Bre'		Bre-Dre */
	fxch	%st(2)	/* Are'		Cre'		Dre'		Bre'		Bre-Dre */
	fstps	Are	/* Cre'		Dre'		Bre'		Bre-Dre */
	fstps	Bre	/* Dre'		Bre'		Bre-Dre */
	fstps	Dre	/* Bre'		Bre-Dre */
	fstps	Cre	/* Bre-Dre */

	/* Now do the imaginaries */
	flds	Aim
	fsubs	Cim	/* Aim-Cim	Bre-Dre */
	flds	Aim
	fadds	Cim	/* Aim+Cim	Aim-Cim		Bre-Dre */
	flds	Bim
	fadds	Dim	/* Bim+Dim	Aim+Cim		Aim-Cim		Bre-Dre */	
	fld	%st(3)	/* Bre-Dre	Bim+Dim		Aim+Cim		Aim-Cim		Bre-Dre */
	fadd	%st(3)	/* Dim'		Bim+Dim		Aim+Cim		Aim-Cim		Bre-Dre */
	fxch	%st(2)	/* Aim+Cim	Bim+Dim		Dim'		Aim-Cim		Bre-Dre */
	fld	%st(0)	/* Aim+Cim	Aim+Cim		Bim+Dim		Dim'		Aim-Cim		Bre-Dre */
	fadd	%st(2)	/* Aim'		Aim+Cim		Bim+Dim		Dim'		Aim-Cim		Bre-Dre */
	fxch	%st(2)	/* Bim+Dim	Aim+Cim		Aim'		Dim'		Aim-Cim		Bre-Dre */
	fsubrp		/* Cim'		Aim'		Dim'		Aim-Cim		Bre-Dre */
	fxch	%st(3)	/* Aim-Cim	Aim'		Dim'		Cim'		Bre-Dre */
	fsub	%st(4)	/* Bim'		Aim'		Dim'		Cim'		Bre-Dre */
	fxch	%st(2)	/* Dim'		Aim'		Bim'		Cim'		Bre-Dre */

	fstps	Dim	/* Aim'		Bim'		Cim'		Bre-Dre */
	fstps	Aim	/* Bim'		Cim'		Bre-Dre */
	fstps	Cim	/* Cim'		Bre-Dre */
	fstps	Bim	/* Bre-Dre */
	ffree	%st(0)

	cmpl	$0,16(%ebp)
	jz done			/* Twiddles are all unity, skip this stuff */

	shll $2,16(%ebp)	/* tstep *= sizeof(float) */
	movl 16(%ebp),%eax	/* %eax = tstep */

	/* Multiply by twiddle factors */
	flds	Cre
	fmuls	COS(%eax)	/* Bre*cos */
	flds	Cim
	fmuls	SIN(%eax)	/* Bim*sin	Bre*cos */
	flds	Cre
	fmuls	SIN(%eax)	/* Bre*sin	Bim*sin		Bre*cos */
	fxch	%st(2)		/* Bre*cos	Bim*sin		Bre*sin */
	fsubp			/* Bre'		Bre*sin */
	fxch	%st(1)		/* Bre*sin	Bre' */
	flds	Cim
	fmuls	COS(%eax)	/* Bim*cos	Bre*sin		Bre' */
	fxch	%st(2)		/* Bre'		Bre*sin		Bim*cos */
	fstps	Cre		/* Bre*sin	Bim*cos */
	faddp			/* Bim' */
	addl %eax,%eax		/* 2*tstep*sizeof(float) */
	flds	Bre		/* Cre		Bim' */
	fxch	%st(1)		/* Bim'		Cre */
	fstps	Cim		/* Cre */
	fmuls	COS(%eax)	/* Cre*cos */
	flds	Bim
	fmuls	SIN(%eax)	/* Cim*sin	Cre*cos */
	flds	Bre
	fmuls	SIN(%eax)	/* Cre*sin	Cim*sin		Cre*cos */
	fxch	%st(2)		/* Cre*cos	Cim*sin		Cre*sin */
	fsubp			/* Cre'		Cre*sin */
	fxch	%st(1)		/* Cre*sin	Cre' */
	flds	Bim
	fmuls	COS(%eax)	/* Cim*cos	Cre*sin		Cre' */
	fxch	%st(2)		/* Cre'		Cre*sin		Cim*cos */
	fstps	Bre		/* Cre*sin	Cim*cos */
	faddp			/* Cim */
	addl	16(%ebp),%eax	/* 3*tstep*sizeof(float) */
	flds	Dre		/* Dre		Cim */
	fxch	%st(1)		/* Cim		Dre */
	fstps	Bim		/* Dre */
	fmuls	COS(%eax)	/* Dre*cos */
	flds	Dim
	fmuls	SIN(%eax)	/* Dim*sin	Dre*cos */
	flds	Dre
	fmuls	SIN(%eax)	/* Dre*sin	Dim*sin		Dre*cos */
	fxch	%st(2)		/* Dre*cos	Dim*sin		Dre*sin */
	fsubp			/* Dre'		Dre*sin */
	fxch	%st(1)		/* Dre*sin	Dre' */
	flds	Dim
	fmuls	COS(%eax)	/* Dim*cos	Dre*sin		Dre' */
	fxch	%st(2)		/* Dre'		Dre*sin		Dim*cos */
	fstps	Dre		/* Dre*sin	Dim*cos */
	faddp			/* Dim */
	fstps	Dim

done:	pop	%edx
	pop	%ecx
	pop	%ebx
	mov	%ebp,%esp
	pop	%ebp
	ret
